/* -*- C++ -*-
 * This file is part of the pSOS simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Interface to the pSOS synchronization objects.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _pSOSSynch_h
#define _pSOSSynch_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "CkSynch.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_PSOS
#include "sodecl.h"

class pSOSNode;
class pSOSTask;

class pSOSSynch : public CkSynchro {

protected:

    int options;

public:

    pSOSSynch(const char *name,
	      InsertMode mode,
	      int options,
	      CkStateExportObject *synchState,
	      CkNode *node);

    pSOSNode *getNode() {
	return (pSOSNode *)getNode();
    }

    int testOption(int option) const {
	return !!(options & option);
    }

    void blockTask(pSOSTask *task,
		   ckticks_t timeout);

    void unblockTask(pSOSTask *task);

    void forgetTask(pSOSTask *task);
};

enum pSOSSemStates {

    pSOSSemIDLE =0,
    pSOSSemPENDED =1,
    pSOSSemPOSTED =2
};

class pSOSSem : public pSOSSynch {

private:

    ckulw_t count;

protected:

    virtual void protoInit();

    static const char *stateArray[3];

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

public:

    pSOSSem(const char *name,
	    CkExtId uname,
	    ckulw_t icount,
	    int options,
	    pSOSNode *node);

    virtual ~pSOSSem();

    int acquire(pSOSTask *task,
		ckticks_t timeout);

    int release();

    int test();

    ckulw_t getCount() {
	return count;
    }

    virtual void forget(CkThread *thread);

    virtual const char *getTypedName();
};

enum pSOSQueueStates {

    pSOSQueueEMPTY =0,
    pSOSQueuePENDED =1,
    pSOSQueuePOSTED =2,
    pSOSQueueFULL =3,
    pSOSQueueJAMMED =4
};

// This flag is cumulative with the pSOS queue options
// defined in psos/psos.h.
#define Q_VARIABLE 0x80000000

struct pSOSMessage {

    ckulw_t length;
    struct pSOSMessage *link;
    ITime time;

    // Start of data area: must be last member.

    union {
	u_long fixbuf[4];
	char data[1];
    } u;
};

class pSOSQueueSlot {

protected:

    pSOSMessage *qhead,
	**qtail;

    ckulw_t count;

public:

    pSOSMessage *pop() {

	pSOSMessage *msg = qhead;

	if (msg)
	    {
	    qhead = msg->link;

	    if (--count == 0)
		qtail = &qhead;
	    }

	return msg;
    }

    void prepend(pSOSMessage *msg) {
	msg->link = qhead;
	qhead = msg;
	count++;
    }

    void append(pSOSMessage *msg) {
	*qtail = msg;
	qtail = &msg->link;
	msg->link = NULL;
	count++;
    }

    pSOSQueueSlot() {
	qhead = NULL;
	qtail = &qhead;
	count = 0;
    }

    virtual ~pSOSQueueSlot() {

	pSOSMessage *msg;

	while ((msg = pop()) != NULL)
	    delete[] (char *)msg;
    }

    ckulw_t getCount() const {
	return count;
    }

    pSOSMessage *first() {
	return qhead;
    }
};

class pSOSQueue : public pSOSSynch {

protected:

    ckulw_t qsize,
	msize;

    pSOSQueueSlot inq;

    pSOSMessage *buildMessage(void *buf,
			      ckulw_t length);

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[5];
	
public:

    pSOSQueue(const char *name,
	      CkExtId uname,
	      ckulw_t qsize,
	      ckulw_t msize,
	      int options,
	      pSOSNode *node);

    virtual ~pSOSQueue();

    int post(void *buf,
	     ckulw_t length);

    int pend(pSOSTask *task,
	     ckticks_t timeout);

    int jam(void *buf,
	    ckulw_t length);

    pSOSMessage *accept();

    ckulw_t broadcast(void *buf,
		      ckulw_t length);

    ckulw_t getMsgCount() {
	return inq.getCount();
    }

    ckulw_t getQueueSize() const {
	return qsize;
    }

    ckulw_t getMsgSize() const {
	return msize;
    }

    void dispose(pSOSMessage *msg) {
	delete[] (char *)msg;
    }

    virtual void forget(CkThread *thread);

    virtual const char *getTypedName();
};

enum pSOSFlagPendModes {

    pSOSFlagPendAND =0,
    pSOSFlagPendOR =1
};

class pSOSFlagGroup : public pSOSSynch {

private:

    ckflags_t fgroup;

protected:

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

public:

    pSOSFlagGroup(const char *name,
		  CkExtId uname,
		  InsertMode mode,
		  CkDomain domain,
		  pSOSNode *node);

    virtual ~pSOSFlagGroup();

    ckflags_t getFlags() const {
	return fgroup;
    }

    int post(ckflags_t mask);

    int pend(pSOSTask *task,
	     ckflags_t mask,
	     pSOSFlagPendModes mode,
	     ckticks_t timeout);

    void accept(ckflags_t *rflags,
		ckflags_t mask);

    ckflags_t poll() const {
	return fgroup;
    }

    virtual void forget(CkThread *thread);

    virtual const char *getTypedName();
};

#endif // !_pSOSSynch_h
