/* -*- C++ -*-
 * This file is part of the pSOS simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Interface to the pSOS task class.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _pSOSTask_h
#define _pSOSTask_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "SxStats.h"
#include "CkThread.h"
#include "CkAlarm.h"
#include "pSOSSynch.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_PSOS
#include "sodecl.h"

class pSOSNode;
class pSOSTask;

enum pSOSTaskStates {

    pSOSTaskHALTED =0,
    pSOSTaskDORMANT =1,
    pSOSTaskBLOCKED =2,
    pSOSTaskREADY =3,
    pSOSTaskEXECUTING =4
};

#define pSOS_NOTEPAD_SIZE 16
#define pSOS_FIRST_SYSREG 8

// Indicates that the user-code entry point has been called
// (i.e. so we can longjmp() to it using restartContext).
#define pSOS_TASK_STARTED CK_THREAD_BIT0

class pSOSTask : public CkThread {

protected:

    void (*startAddr)(u_long,u_long,u_long,u_long);

    pSOSArgs startArgs;

    ckflags_t startMode;

    int startPri;

    int startTaskIL;

    // A SIMEX exception is used to restart a pSOSTask.
    SxThreadContext restartContext;

    pSOSFlagGroup *eventGroup;

    union {

	pSOSMessage *msg;

	struct {
	    pSOSFlagPendModes pmode;
	    ckflags_t flags;
	} fgroup;

	struct {
	    ckulw_t size;
	    void *addr;
	} segment;

    } pendedResourceArgs;

    ITime lastStateChangeTime;

    SxHistogram *stateDistribs[3];

    u_long notepad[pSOS_NOTEPAD_SIZE];

    void logTaskState(pSOSTaskStates s);

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    void kroot(run)();

public:

    static const char *stateArray[5];

    pSOSTask(const char *name,
	     CkExtId uname,
	     int priority,
	     CkDomain domain,
	     pSOSNode *node,
	     unsigned stackSize =0);

    void start(void (*startAddr)(u_long,u_long,u_long,u_long),
	       ckflags_t mode,
	       int il,
	       pSOSArgs *targs);

    void restart(pSOSArgs *args);

    virtual ~pSOSTask();

    pSOSNode *getNode() {
	return (pSOSNode *)CkThread::getNode();
    }

    void setRegister(int regNum, u_long value) {
	notepad[regNum] = value;
    }

    u_long getRegister(int regNum) const {
	return notepad[regNum];
    }

    void msgSet(pSOSMessage *msg) {
	pendedResourceArgs.msg = msg;
    }

    pSOSMessage *msgGet() {
	return pendedResourceArgs.msg;
    }

    void flagValSet(ckflags_t f) {
	pendedResourceArgs.fgroup.flags = f;
    }

    ckflags_t flagValGet() {
	return pendedResourceArgs.fgroup.flags;
    }

    void flagModeSet(pSOSFlagPendModes m) {
	pendedResourceArgs.fgroup.pmode = m;
    }

    pSOSFlagPendModes flagModeGet() {
	return pendedResourceArgs.fgroup.pmode;
    }

    void segSizeSet(ckulw_t sz) {
	pendedResourceArgs.segment.size = sz;
    }

    ckulw_t segSizeGet() {
	return pendedResourceArgs.segment.size;
    }

    void segAddrSet(void *addr) {
	pendedResourceArgs.segment.addr = addr;
    }

    void *segAddrGet() {
	return pendedResourceArgs.segment.addr;
    }

    int postEvents(ckflags_t events) {
	return eventGroup->post(events);
    }

    int pendEvents(ckflags_t events,
		   pSOSFlagPendModes mode,
		   ckticks_t timeout) {
	return eventGroup->pend(this,events,mode,timeout);
    }

    void acceptEvents(ckflags_t *flags,
		      ckflags_t mask) {
	eventGroup->accept(flags,mask);
    }

    ckflags_t pollEvents() const {
	return eventGroup->poll();
    }

    virtual void body();

    virtual void setReady();

    virtual void setHalted();

    virtual void setDelayed();

    virtual void setPending();

    virtual void setRunning();

    virtual void setSuspended();

    virtual void setDormant();

    virtual void (*getStartAddr())(...);

    virtual void startAsr(cksigmask_t sigs);

    virtual void protoInit();
};

// T_NOPREEMPT/T_SLICE/T_NOASR <=> CKLOCK/CKRRB/CKASDI
#define psos2vrtosMode(mode) (((mode) << 13) & CK_MODE_BITS)
#define vrtos2psosMode(mode) ((mode) >> 13)

MakeGList(pSOSTask);

class pSOSAlarm : public CkAlarm {

protected:

    pSOSTask *clientTask;

public:

    pSOSAlarm(pSOSTask *task, CkDateVal dv);

    virtual void timeout(CkObject *client);
};

class pSOSTimer : public CkAlarm {

private:

    pSOSTask *clientTask;

    ckflags_t events;

public:

    pSOSTimer(pSOSTask *task,
	      CkDateVal dv,
	      ckflags_t events);

    pSOSTimer(pSOSTask *task,
	      ckticks_t initialTicks,
	      ckticks_t reloadTicks,
	      ckflags_t events);

    virtual ~pSOSTimer();

    virtual void timeout(CkObject *client);
};

#endif // !_pSOSTask_h
