/*
 * This file is part of the pSOS simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the pSOS+m target abstraction.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation "pSOSNode.h"
#pragma implementation "pSOSProto.h"
#endif // __GNUG__
#include <config.h>
#include "CkManager.h"
#include "pSOSNode.h"
#include "pSOSTask.h"
#include "pSOSMemory.h"
#include "pSOSProto.h"
#include "pSOSTrace.h"

pSOSNode::pSOSNode (const char *_nodeName,
		    CkNodeParms *_nparms) :
    CkNode("pSOS+m",
	   _nodeName,
	   pSOS_MAGIC,
	   _nparms)
{
    utask = 0;
    uqueue = 0;
    usema4 = 0;
    umsgbuf = 0;
    utimer = 0;
    ulocobj = 0;
    uglbobj = 0;
    uagent = 0;

    fmsgbuf = nmsgbuf;	// Count of available message buffers in shared pool.

    // Configure and start the real-time clock. A value of zero for the
    // ticks2sec configuration value disables the RTC for the node.
    // A null handler also prevents the RTC installation.

    if (getTicks2Sec() > 0 && _nparms->rtcIntrHandler != NULL)
	{
	// Create the interrupt object underlying the system timer.

	rtc = createInterrupt(_nparms->rtcIntrVector,
			      _nparms->rtcIntrLevel,
			      (ckisr_t *)_nparms->rtcIntrHandler,
			      NULL,
			      0,
			      CK_RTC_NAME);

	rtcProgram(getTicks2Sec());
	}

    setTraceInterface(new pSOSTraceInterface(this));
}

CkInterrupt *pSOSNode::createInterrupt (ckvector_t vector,
					cklevel_t level,
					ckisr_t *isr,
					ckdsr_t *dsr,
					void *cookie,
					const char *symbolicName)
{
    return new pSOSInterrupt(vector,
			     level,
			     (void (*)(void))isr,
			     symbolicName,
			     this);
}

void pSOSNode::kinit(coldInit) ()

{
    CkNode::kinit(coldInit)();

    // Create region #0 (rn0len can't be null).

    char *rn0addr = new char[rn0len];

    if (!rn0addr)
	CkManager::This->fatal("Cannot allocate %u bytes for region #0\n",rn0len);

    region0 = new pSOSRegion("RN#0",
			     "RN#0",
			     rn0addr,
			     rn0len,
			     rn0usize,
			     False,
			     FIFO,
			     this);
}

void pSOSNode::kinit(warmInit) ()

{
    CkNode::kinit(warmInit)();

    pSOSTask *rootTask = new pSOSTask("ROOT",
				      "ROOT",
				      rootTaskPri,
				      CkLocalDomain,
				      this,
				      32768);

    rootTask->start((void (*)(u_long,u_long,u_long,u_long))rootTaskEntry,
		     rootTaskMode,
		     0,
		     0);
}

void pSOSNode::protoInit ()

{
    pSOSNodeExportMessage nem(this);
    protoExport(&nem,sizeof(nem));
}

ckticks_t pSOSNode::rrGetCredit () {

    return ticks2slice;
}

void (*pSOSNode::getRootEntry ()) (...) {

    return rootTaskEntry;
}

void pSOSNode::kcout(fireCallout) (CkEventKind event, CkObject *object)

{
    CkNode::kcout(fireCallout)(event,object);

    CkThread *thread = (CkThread *)object;
    ckhandle_t srcid = thread->getCkId(), dstid = CkNoId;

    if (thread->getCoSource())
	dstid = thread->getCoSource()->getCkId();

    switch (event)
	{
	case XTStartEvent:

	    if (startCO)
		startCO(srcid,dstid);

	    break;

	case XTSwitchEvent:

	    if (switchCO)
		switchCO(srcid,dstid);

	    break;

	case XTDeleteEvent:

	    if (deleteCO)
		deleteCO(srcid,dstid);

	    break;

	default: ;
	}
}

void pSOSNode::nameObject (CkObject *object,
			   int type,
			   CkDomain domain,
			   CkExtId uname)
{
    int seqno = allocID();

    CkId id(getNid(),type,domain == CkGlobalDomain,seqno);

    // If exported, this object must be inserted into all known
    // pSOS nodes' GOT (including this one).

    if (id.isGlobal())
	{
	uglbobj++;	// Track KC_NGLBOBJ value.

	pSOSNodeIterator it(CkNodeList);
	pSOSNode *node;

	while ((node = it.next()) != NULL)
	    {
	    if (node->getNodeType() != pSOS_MAGIC)
		continue;

	    // Remove id from the deleted object search tree,
	    // then insert it into the active object search tree.

	    node->unmapObject(id,type + 1,domain);
	    node->mapObject(id,type,domain,object);
	    }
	}
    else
	{
	ulocobj++;	// Track KC_NLOCOBJ value.
	unmapObject(id,type + 1,domain);
	mapObject(id,type,domain,object);
	}

    object->setCkId(id);
    object->setCkExtId(uname);

    tellNameServer(id,CK_MSGPORT_NAMEOBJ);
}

void pSOSNode::unnameObject (CkObject *object)

{
    // If exported, this object must be removed from all known
    // pSOS nodes' GOT (including this one).

    if (object->getCkId().isGlobal())
	{
	uglbobj++;

	pSOSNodeIterator it(CkNodeList);
	pSOSNode *node;

	while ((node = it.next()) != NULL)
	    {
	    if (node->getNodeType() != pSOS_MAGIC)
		continue;

	    // Remove the object id from the active object search tree,
	    // then insert it into the deleted object search tree.

	    node->unmapObject(object->getCkId(),
			      object->getCkType(),
			      CkGlobalDomain);

	    node->mapObject(object->getCkId(),
			    object->getCkType() + 1,
			    CkGlobalDomain,
			    object);
	    }
	}
    else
	{
	ulocobj--;

	unmapObject(object->getCkId(),
		    object->getCkType(),
		    CkLocalDomain);

	mapObject(object->getCkId(),
		  object->getCkType() + 1,
		  CkLocalDomain,
		  object);
	}

    if (object->getCkId().getType() == pSOS_TIMER_INODE)
	utimer--;	// Special case to ease bookkeeping.

    tellNameServer(object->getCkId(),CK_MSGPORT_UNNAMEOBJ);
}

CkObject *pSOSNode::findActiveObject (CkId id)

{
    CkDomain domain = id.isGlobal() ? CkGlobalDomain : CkLocalDomain;
    return findObject(id,id.getType(),domain);
}

CkObject *pSOSNode::findActiveObject (CkExtId uname, int nid, int type)

{
    CkObject *object;
    CkInode *inode;

    if (nid > 0)
	{
	// Search into the given node's GOT.

	List *l = CkNodeTable[nid - 1]->getObjectList(type,CkGlobalDomain);

	for (inode = (CkInode *)l->first();
	     inode; inode = (CkInode *)inode->next())
	    {
	    object = (CkObject *)*inode;

	    if (object->getCkExtId() == uname)
		return object;
	    }
	}
    else
	{
	// Search into the current node's LOT, then into all nodes' GOT
	// if not previously found, starting from the current node's.

	List *l = getObjectList(type,CkLocalDomain);

	for (inode = (CkInode *)l->first();
	     inode; inode = (CkInode *)inode->next())
	    {
	    object = (CkObject *)*inode;

	    if (object->getCkExtId() == uname)
		return object;
	    }

	int rtnid = getNid();

	do
	    {
	    if (CkNodeTable[rtnid]->getNodeType() == pSOS_MAGIC)
		{
		List *l = CkNodeTable[rtnid]->getObjectList(type,CkGlobalDomain);
		     
		for (inode = (CkInode *)l->first();
		     inode; inode = (CkInode *)inode->next())
		    {
		    object = (CkObject *)*inode;
		    
		    if (object->getCkExtId() == uname)
			return object;
		    }
		}
	    }
	while ((++rtnid %= CkNodeList.getCount()) != getNid());
	}

    return NULL;
}

int pSOSNode::isDeletedObject (CkId id)

{
    CkDomain domain = id.isGlobal() ? CkGlobalDomain : CkLocalDomain;
    return !!findObject(id,id.getType() + 1,domain);
}

int pSOSNode::checkpSOSNodeID (int nid)

{
    if (nid != 0 &&
	((nid < 0 ||
	  (unsigned)nid > CkNodeList.getCount() ||
	  CkNodeTable[nid - 1]->getNodeType() != pSOS_MAGIC)))
	 return -1;

    return 0;
}

void pSOSNode::dispatchAsyncErr (int rc)

{
    if (asyncHandler)
	asyncHandler(rc);
    else
	{
	if (fatalHandler)
	    fatalHandler(rc);

	CkManager::This->fatal("Error from asynchronous service: status=%d\n",rc);
	}
}

int pSOSNode::kdoor(tcreate) (CkExtId uname,
			      u_long pri,
			      unsigned stackSize,
			      ckflags_t flags,
			      CkId *tidp)
{
    mustBeSynch();

    utask++;	// Track KC_NTASK value.

    CkDomain domain = (flags & T_GLOBAL) ? CkGlobalDomain : CkLocalDomain;

    pSOSTask *task = new pSOSTask(CkNameOf(uname),
				  uname,
				  pri,
				  domain,
				  this,
				  stackSize);
    *tidp = task->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(tident) (CkExtId uname,
			     u_long nid,
			     CkId *tidp)
{
    mustBeSynch();

    if (uname == CkNoExtId)
	*tidp = pSOSRunningTask->getCkId();
    else
	{
	if (checkpSOSNodeID(nid) < 0)
	    return ERR_NODENO;

	CkObject *object = findActiveObject(uname,nid,pSOS_TASK_INODE);
	
	if (!object)
	    return ERR_OBJNF;

	*tidp = object->getCkId();
	}

    return SUCCESS;
}

int pSOSNode::kdoor(tstart) (CkId tid,
			     ckflags_t mode,
			     void (*startAddr)(u_long,u_long,u_long,u_long),
			     pSOSArgs *targs)
{
    mustBeSynch();

    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;	// Not a task id.

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    if (tid.getNode() != getNid())
	return ERR_ILLRSC;	// Task belongs to a remote node.

    if (!task->tstStatus(CKWAIT))
	return ERR_ACTIVE;	// Task already started.

    int il = (int)((mode >> 8) & 0x7);	// Fetch interrupt level.
    ckflags_t rtmode = psos2vrtosMode(mode); // Convert modebits pSOS => VRTOS.

    task->start(startAddr,rtmode,il,targs); // Calls CkNode::reched().

    return SUCCESS;
}

int pSOSNode::kdoor(trestart) (CkId tid,
			       pSOSArgs *targs)
{
    mustBeSynch();

    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    if (tid.getNode() != getNid())
	return ERR_ILLRSC;

    if (task->tstStatus(CKWAIT))
	return ERR_NACTIVE;

    task->restart(targs);

    resched();

    return SUCCESS;
}

int pSOSNode::kdoor(tdelete) (CkId tid)

{
    mustBeSynch();

    if ((CkHandle)tid == 0)
	deleteThread(pSOSRunningTask);
    else
	{
	if (tid.getType() != pSOS_TASK_INODE)
	    return ERR_OBJTYPE;

	pSOSTask *task = (pSOSTask *)findActiveObject(tid);

	if (!task)
	    return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

	if (tid.getNode() != getNid())
	    return ERR_ILLRSC;

	deleteThread(task);
	}

    utask--;

    resched();

    return SUCCESS;
}

int pSOSNode::kdoor(tsuspend) (CkId tid)

{
    mustBeSynch();

    if ((CkHandle)tid == 0)
	{
	suspendThread(pSOSRunningTask,CKSUSP);
	resched();
	}
    else
	{
	if (tid.getType() != pSOS_TASK_INODE)
	    return ERR_OBJTYPE;

	pSOSTask *task = (pSOSTask *)findActiveObject(tid);

	if (!task)
	    return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

	if (task->tstStatus(CKSUSP))
	    return ERR_SUSP;	// Task already suspended.

	CkNode *node = CkNodeTable[tid.getNode()];

	node->suspendThread(task,CKSUSP);

	node->resched();
	}

    return SUCCESS;
}

int pSOSNode::kdoor(tresume) (CkId tid)

{
    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(tid);

    if (!task->tstStatus(CKSUSP))
	return ERR_NOTSUSP;	// Task not suspended.

    CkNode *targetNode = CkNodeTable[tid.getNode()];

    targetNode->readyThread(task,CKSUSP);

    targetNode->resched();

    return SUCCESS;
}

int pSOSNode::kdoor(tsetpri) (CkId tid,
			      u_long newPri,
			      u_long *oldPri)
{
    mustBeSynch();

    pSOSTask *task;

    if ((CkHandle)tid == 0)
	task = pSOSRunningTask;
    else
	{
	if (tid.getType() != pSOS_TASK_INODE)
	    return ERR_OBJTYPE;

	task = (pSOSTask *)findActiveObject(tid);

	if (!task)
	    return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;
	}

    *oldPri = task->getPri();

    if (newPri != 0)
	{
	if (newPri < 1 || newPri > 255)
	    return ERR_SETPRI;

	if (newPri != *oldPri)
	    {
	    CkNode *targetNode = CkNodeTable[task->getNode()->getNid()];
	    targetNode->reniceThread(task,newPri);
	    targetNode->resched();
	    }
	}

    return SUCCESS;
}

int pSOSNode::kdoor(tmode) (ckflags_t mask,
			    ckflags_t newMode,
			    ckflags_t *oldMode)
{
    mustBeSynch();

    *oldMode = vrtos2psosMode(pSOSRunningTask->tstStatus(CK_MODE_BITS));
    *oldMode |= ((pSOSRunningTask->getThreadIL() & 0x7) << 8);

    if (mask != 0)
	{
	ckflags_t rtmask = psos2vrtosMode(mask);
	pSOSRunningTask->clrStatus(rtmask);

	ckflags_t rtmode = psos2vrtosMode(newMode);
	pSOSRunningTask->setStatus(rtmode);

	resched();
	}

    return SUCCESS;
}

int pSOSNode::kdoor(tgetreg) (CkId tid,
			      u_long regNum,
			      u_long *regValuep)
{
    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(tid);

    if (regNum < 0 || regNum >= pSOS_NOTEPAD_SIZE)
	return ERR_REGNUM;

    *regValuep = task->getRegister(regNum);

    return SUCCESS;
}

int pSOSNode::kdoor(tsetreg) (CkId tid,
			      u_long regNum,
			      u_long regValue)
{
    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(tid);

    if (regNum < 0 || regNum >= pSOS_NOTEPAD_SIZE)
	return ERR_REGNUM;

    if (regNum >= pSOS_FIRST_SYSREG)
	CkManager::This->warning("use of reserved notepad register #%d",regNum);

    task->setRegister(regNum,regValue);

    return SUCCESS;
}

int pSOSNode::kdoor(rncreate) (CkExtId uname,
			       void *rnaddr,
			       ckulw_t rnSz,
			       ckulw_t unitSz,
			       ckflags_t flags,
			       CkId *rnidp,
			       ckulw_t *allocSzp)
{
    mustBeSynch();

    if (((ckulw_t)rnaddr % sizeof(ckulw_t)) != 0)
	return ERR_RNADDR;	// Memory address not aligned.

    if (unitSz < 16)
	return ERR_UNITSIZE;

    if (unitSz & (unitSz - 1))
	return ERR_UNITSIZE;	// Not a power of two.

    if (rnSz <= pSOS_RNCB(rnSz,CkAlignLW(unitSz)))
	return ERR_TINYRN;

    if (rnSz / CkAlign(unitSz,sizeof(ckulw_t)) > 32767)
	return ERR_TINYUNIT;

    pSOSRegion *region = new pSOSRegion(CkNameOf(uname),
					uname,
					rnaddr,
					rnSz,
					unitSz,
					!!(flags & RN_DEL),
					(flags & RN_PRIOR) ? PRDN : FIFO,
					this);
    *rnidp = region->getCkId();

    *allocSzp = region->getMaxAlloc();

    return SUCCESS;
}

int pSOSNode::kdoor(rnident) (CkExtId uname,
			      CkId *rnid)
{
    mustBeSynch();

    // Regions are strictly local objects: search in the LOT only.
    CkObject *object = findActiveObject(uname,-1,pSOS_REGION_INODE);
	
    if (!object)
	return ERR_OBJNF;

    *rnid = object->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(rndelete) (CkId rnid)

{
    mustBeSynch();

    if ((CkHandle)rnid == 0)
	return ERR_OBJID;

    if (rnid.getType() != pSOS_REGION_INODE)
	return ERR_OBJTYPE;

    pSOSRegion *region = (pSOSRegion *)findActiveObject(rnid);

    if (!region)
	return isDeletedObject(rnid) ? ERR_OBJDEL :  ERR_OBJID;

    if (region->vetoDestroy())
	return ERR_SEGINUSE;	// Deletion denied.

    int rc = region->getPendCount() > 0 ? ERR_TATRNDEL : SUCCESS;

    delete region;

    if (rc == ERR_TATRNDEL)
	resched();
    
    return rc;
}

int pSOSNode::kdoor(rngetseg) (CkId rnid,
			       ckulw_t segSz,
			       ckflags_t flags,
			       ckticks_t timeout,
			       void **segp)
{
    mustBeSynch();

    pSOSRegion *region;

    if ((CkHandle)rnid == 0)
	region = region0;
    else if (rnid.getType() != pSOS_REGION_INODE)
	return ERR_OBJTYPE;
    else
	{
	region = (pSOSRegion *)findActiveObject(rnid);

	if (!region)
	    return isDeletedObject(rnid) ? ERR_OBJDEL :  ERR_OBJID;
	}

    if (segSz > region->getMaxSegSize())
	return ERR_TOOBIG;

    if (flags & RN_NOWAIT)
	{
	if (region->grabSegment(pSOSRunningTask,segSz) == CK_SYNCH_NOWAIT)
	    return ERR_NOSEG;
	}
    else if (region->allocSegment(pSOSRunningTask,segSz,timeout) == CK_SYNCH_WAIT)
	{
	resched();

	if (pSOSRunningTask->tstStatus(CKTIMEO))
	    return ERR_TIMEOUT;

	if (pSOSRunningTask->tstStatus(CKRMID))
	    return ERR_RNKILLD;
	}

    *segp = pSOSRunningTask->segAddrGet();

    return SUCCESS;
}

int pSOSNode::kdoor(rnretseg) (CkId rnid,
			       void *seg)
{
    mustBeSynch();

    pSOSRegion *region;

    if ((CkHandle)rnid == 0)
	region = region0;
    else if (rnid.getType() != pSOS_REGION_INODE)
	return ERR_OBJTYPE;
    else
	{
	region = (pSOSRegion *)findActiveObject(rnid);

	if (!region)
	    return isDeletedObject(rnid) ? ERR_OBJDEL :  ERR_OBJID;
	}

    CkSegmentStatus segStatus = region->testSeg(seg);

    switch (segStatus)
	{
	case CK_MEMORY_BAD:

	    return ERR_SEGADDR;

	case CK_MEMORY_NIR:

	    return ERR_NOTINRN;

	case CK_MEMORY_FREE:

	    return ERR_SEGFREE;

	default:

	    if (region->freeSegment(seg) == CK_SYNCH_RESCHED)
		resched();
	}

    return SUCCESS;
}

int pSOSNode::kdoor(ptcreate) (CkExtId uname,
			       void *paddr,
			       void *laddr, // logical address: n/a under Ck
			       ckulw_t psize,
			       ckulw_t bsize,
			       ckflags_t flags,
			       CkId *ptidp,
			       ckulw_t *nbufp)
{
    mustBeSynch();

    if ((ckulw_t)paddr % sizeof(void *))
	return ERR_PTADDR;

    if (bsize < 4)
	return ERR_BUFSIZE;

    if (bsize & (bsize - 1))
	return ERR_BUFSIZE;	// Not a power of two.

    if (psize < pSOS_PTCB(psize,CkAlignLW(bsize)))
	return ERR_TINYPT;

    pSOSPartition *part = new pSOSPartition(CkNameOf(uname),
					    uname,
					    paddr,
					    psize,
					    bsize,
					    !!(flags & PT_DEL),
					    (flags & PT_GLOBAL) ? CkGlobalDomain : CkLocalDomain,
					    this);
    *ptidp = part->getCkId();

    *nbufp = part->getMaxBlks();

    return SUCCESS;
}

int pSOSNode::kdoor(ptident) (CkExtId uname,
			      u_long nid,
			      CkId *ptidp)
{
    mustBeSynch();

    if (checkpSOSNodeID(nid) < 0)
	return ERR_NODENO;

    CkObject *object = findActiveObject(uname,nid,pSOS_PART_INODE);
	
    if (!object)
	return ERR_OBJNF;

    *ptidp = object->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(ptdelete) (CkId ptid)

{
    mustBeSynch();

    if (ptid.getType() != pSOS_PART_INODE)
	return ERR_OBJTYPE;

    pSOSPartition *part = (pSOSPartition *)findActiveObject(ptid);

    if (!part)
	return isDeletedObject(ptid) ? ERR_OBJDEL :  ERR_OBJID;

    if (ptid.getNode() != getNid())
	return ERR_ILLRSC;

    if (part->vetoDestroy())
	return ERR_BUFINUSE;

    delete part;

    return SUCCESS;
}

int pSOSNode::kdoor(ptgetbuf) (CkId ptid,
			       void **bufAddrp)
{
    if (ptid.getType() != pSOS_PART_INODE)
	return ERR_OBJTYPE;

    pSOSPartition *part = (pSOSPartition *)findActiveObject(ptid);

    if (!part)
	return isDeletedObject(ptid) ? ERR_OBJDEL :  ERR_OBJID;

    mustBeLocalOrSynch(ptid);

    *bufAddrp = part->gblock();

    return *bufAddrp ? SUCCESS : ERR_NOBUF;
}

int pSOSNode::kdoor(ptretbuf) (CkId ptid,
			       void *bufAddr)
{
    if (ptid.getType() != pSOS_PART_INODE)
	return ERR_OBJTYPE;

    pSOSPartition *part = (pSOSPartition *)findActiveObject(ptid);

    if (!part)
	return isDeletedObject(ptid) ? ERR_OBJDEL :  ERR_OBJID;

    mustBeLocalOrSynch(ptid);

    int rc = part->rblock(bufAddr);

    return rc > 0 ? SUCCESS : rc == 0 ? ERR_BUFFREE : ERR_BUFADDR;
}

int pSOSNode::kdoor(qcreate) (CkExtId uname,
			      ckulw_t count,
			      ckflags_t flags,
			      CkId *qidp)
{
    mustBeSynch();

    uqueue++; // Track KC_NQUEUE value.

    if (flags & Q_LIMIT)
	 {
	 if (!(flags & Q_PRIBUF))
	     {
	     if (count > fmsgbuf)
		 return ERR_NOMGB;

	     holdMsgBufs(count);
	     }
	 }
    else
	count = ~0; // No queue limit.

    pSOSQueue *queue = new pSOSQueue(CkNameOf(uname),
				     uname,
				     count,
				     sizeof(u_long[4]),
				     flags,
				     this);
    *qidp = queue->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(qvcreate) (CkExtId uname,
			       ckflags_t flags,
			       ckulw_t count,
			       ckulw_t msglen,
			       CkId *qidp)
{
    mustBeSynch();

    uqueue++;	// Track KC_NQUEUE value.

    if (flags & Q_LIMIT)
	 {
	 if (!(flags & Q_PRIBUF))
	     {
	     if (count > fmsgbuf)
		 return ERR_NOMGB;

	     holdMsgBufs(count);
	     }
	 }
    else
	count = ~0; // No queue limit.

    pSOSQueue *queue = new pSOSQueue(CkNameOf(uname),
				     uname,
				     count,
				     msglen,
				     flags|Q_VARIABLE,
				     this);
    *qidp = queue->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(qident) (CkExtId uname,
			     u_long nid,
			     CkId *qidp)
{
    mustBeSynch();

    if (checkpSOSNodeID(nid) < 0)
	return ERR_NODENO;

    CkObject *object = findActiveObject(uname,nid,pSOS_QUEUE_INODE);
	
    if (!object)
	return ERR_OBJNF;

    *qidp = object->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(qvident) (CkExtId uname,
			      u_long nid,
			      CkId *qidp) {

    return kdoor(qident)(uname,nid,qidp);
}

int pSOSNode::kdoor(qdelete) (CkId qid)

{
    mustBeSynch();

    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (qid.getNode() != getNid())
	return ERR_ILLRSC;

    if (queue->testOption(Q_VARIABLE))
	return ERR_VARQ;
    
    uqueue--;

    int rc;

    if (queue->getPendCount() > 0)
	rc = ERR_TATQDEL;
    else if (queue->getMsgCount() > 0)
	rc = ERR_MATQDEL;
    else
	rc = SUCCESS;

    delete queue;

    if (rc == ERR_TATQDEL)
	// Since remote tasks may be pending on this queue, start
	// the multi-processor rescheduling procedure.
	multiProcSched(pSOS_MAGIC);

    return rc;
}

int pSOSNode::kdoor(qvdelete) (CkId qid)

{
    mustBeSynch();

    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (!queue->testOption(Q_VARIABLE))
	return ERR_NOTVARQ;
    
    if (qid.getNode() != getNid())
	return ERR_ILLRSC;

    uqueue--;

    int rc;

    if (queue->getPendCount() > 0)
	rc = ERR_TATQDEL;
    else if (queue->getMsgCount() > 0)
	rc = ERR_MATQDEL;
    else
	rc = SUCCESS;

    delete queue;

    if (rc == ERR_TATQDEL)
	// Since remote tasks may be pending on this queue, start
	// the multi-processor rescheduling procedure.
	multiProcSched(pSOS_MAGIC);

    return rc;
}

int pSOSNode::kdoor(qsend) (CkId qid,
			    void *buf)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(qid);
    
    if (queue->testOption(Q_VARIABLE))
	return ERR_VARQ;
    
    int rc = queue->post(buf,sizeof(u_long[4]));

    if (rc == CK_SYNCH_ERROR)
	return ERR_QFULL;
    
    if (rc == CK_SYNCH_NOWAIT)
	return ERR_NOMGB;

    if (rc == CK_SYNCH_RESCHED)
	multiProcSched(pSOS_MAGIC);

    return SUCCESS;
}

int pSOSNode::kdoor(qasend) (CkId qid,
			     void *buf)
{
    mustBeSynch();
    
    int rc = kdoor(qsend)(qid,buf);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(qvsend) (CkId qid,
			     void *buf,
			     ckulw_t length)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (!queue->testOption(Q_VARIABLE))
	return ERR_NOTVARQ;

    mustBeLocalOrSynch(qid);
    
    if (length > queue->getMsgSize())
	return ERR_MSGSIZ;	// Message too large.

    int rc = queue->post(buf,length);

    if (rc == CK_SYNCH_ERROR)
	return ERR_QFULL;
    
    if (rc == CK_SYNCH_NOWAIT)
	return ERR_NOMGB;

    if (rc == CK_SYNCH_RESCHED)
	multiProcSched(pSOS_MAGIC);

    return SUCCESS;
}

int pSOSNode::kdoor(qavsend) (CkId qid,
			      void *buf,
			      ckulw_t length)
{
    mustBeSynch();
    
    int rc = kdoor(qvsend)(qid,buf,length);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(qurgent) (CkId qid,
			      void *buf)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (queue->testOption(Q_VARIABLE))
	return ERR_VARQ;

    mustBeLocalOrSynch(qid);
    
    int rc = queue->jam(buf,sizeof(u_long[4]));

    if (rc == CK_SYNCH_ERROR)
	return ERR_QFULL;
    
    if (rc == CK_SYNCH_NOWAIT)
	return ERR_NOMGB;

    if (rc == CK_SYNCH_RESCHED)
	multiProcSched(pSOS_MAGIC);

    return SUCCESS;
}

int pSOSNode::kdoor(qaurgent) (CkId qid,
			       void *buf)
{
    mustBeSynch();
		 
    int rc = kdoor(qurgent)(qid,buf);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(qvurgent) (CkId qid,
			       void *buf,
			       ckulw_t length)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (!queue->testOption(Q_VARIABLE))
	return ERR_NOTVARQ;

    mustBeLocalOrSynch(qid);
    
    if (length > queue->getMsgSize())
	return ERR_MSGSIZ;

    int rc = queue->jam(buf,length);

    if (rc == CK_SYNCH_ERROR)
	return ERR_QFULL;
    
    if (rc == CK_SYNCH_NOWAIT)
	return ERR_NOMGB;

    if (rc == CK_SYNCH_RESCHED)
	multiProcSched(pSOS_MAGIC);

    return SUCCESS;
}

int pSOSNode::kdoor(qavurgent) (CkId qid,
				void *buf,
				ckulw_t length)
{
    mustBeSynch();
    
    int rc = kdoor(qvurgent)(qid,buf,length);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(qbroadcast) (CkId qid,
				 void *buf,
				 ckulw_t *countp)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (queue->testOption(Q_VARIABLE))
	return ERR_VARQ;

    mustBeLocalOrSynch(qid);
    
    ckulw_t qpended = queue->getPendCount();
    ckulw_t qposted = queue->broadcast(buf,sizeof(u_long[4]));

    *countp = qposted;

    if (qposted > 0)
	multiProcSched(pSOS_MAGIC);

    return qposted < qpended ? ERR_QFULL : SUCCESS;
}

int pSOSNode::kdoor(qvbroadcast) (CkId qid,
				  void *buf,
				  ckulw_t length,
				  ckulw_t *countp)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (!queue->testOption(Q_VARIABLE))
	return ERR_NOTVARQ;

    if (length > queue->getMsgSize())
	return ERR_MSGSIZ;	// Message too large.

    mustBeLocalOrSynch2(qid);
    
    ckulw_t qpended = queue->getPendCount();
    ckulw_t qposted = queue->broadcast(buf,length);

    *countp = qposted;

    if (qposted > 0)
	multiProcSched(pSOS_MAGIC);

    return qposted < qpended ? ERR_QFULL : SUCCESS;
}

int pSOSNode::kdoor(qreceive) (CkId qid,
			       ckflags_t flags,
			       ckticks_t timeout,
			       u_long buf[4])
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (queue->testOption(Q_VARIABLE))
	return ERR_VARQ;

    pSOSMessage *msg;
    
    if (flags & Q_NOWAIT)
	{
	mustBeLocalOrSynch(qid);
	
	msg = queue->accept();
	
	if (!msg)
	    return ERR_NOMSG;
	}
    else
	{
	mustBeSynch();

	if (queue->pend(pSOSRunningTask,timeout) == CK_SYNCH_WAIT)
	    {
	    // Since the caller is the running task, call the
	    // local node's rescheduling proc.

	    resched();

	    if (pSOSRunningTask->tstStatus(CKRMID))
		return ERR_QKILLD; // Queue deleted while pending.

	    if (pSOSRunningTask->tstStatus(CKTIMEO))
		return ERR_TIMEOUT; // Timeout.
	    }

	msg = pSOSRunningTask->msgGet();
	pSOSRunningTask->msgSet(NULL);
	}

    memcpy(buf,msg->u.data,sizeof(buf));
    queue->dispose(msg);

    return SUCCESS;
}

int pSOSNode::kdoor(qvreceive) (CkId qid,
				ckflags_t flags,
				ckticks_t timeout,
				void *buf,
				ckulw_t length,
				ckulw_t *msize)
{
    if (qid.getType() != pSOS_QUEUE_INODE)
	return ERR_OBJTYPE;

    pSOSQueue *queue = (pSOSQueue *)findActiveObject(qid);

    if (!queue)
	return isDeletedObject(qid) ? ERR_OBJDEL : ERR_OBJID;

    if (!queue->testOption(Q_VARIABLE))
	return ERR_NOTVARQ;

    if (length < queue->getMsgSize())
	return ERR_BUFSIZ;	// Buffer too small.
    
    pSOSMessage *msg;
    
    if (flags & Q_NOWAIT)
	{
	msg = queue->accept();
	
	if (!msg)
	    return ERR_NOMSG;
	}
    else
	{
	mustBeSynch();

	if (queue->pend(pSOSRunningTask,timeout) == CK_SYNCH_WAIT)
	    {
	    // Since the caller is the running task, call the
	    // local node's rescheduling proc.

	    resched();

	    if (pSOSRunningTask->tstStatus(CKRMID))
		return ERR_QKILLD; // Queue deleted while pending.

	    if (pSOSRunningTask->tstStatus(CKTIMEO))
		return ERR_TIMEOUT; // Timeout.
	    }

	msg = pSOSRunningTask->msgGet();
	pSOSRunningTask->msgSet(NULL);
	}

    *msize = Min(length,msg->length);
    memcpy(buf,msg->u.data,*msize);
    queue->dispose(msg);

    return SUCCESS;
}

int pSOSNode::kdoor(evsend) (CkId tid,
			     ckflags_t events)
{
    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(tid);

    if (task->postEvents(events) == CK_SYNCH_RESCHED)
	{
	CkNode *targetNode = CkNodeTable[tid.getNode()];
	targetNode->resched();
	}

    return SUCCESS;
}

int pSOSNode::kdoor(evasend) (CkId tid,
			      ckflags_t events)
{
    mustBeSynch();
    
    int rc = kdoor(evsend)(tid,events);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(evreceive) (ckflags_t events,
				ckflags_t flags,
				ckticks_t timeout,
				ckflags_t *eventsp)
{
    mustBeSynch();

    if (events == 0)
	{
	*eventsp = pSOSRunningTask->pollEvents();
	return SUCCESS;
	}
	    
    pSOSFlagPendModes mode = (flags & EV_ANY) ? pSOSFlagPendOR : pSOSFlagPendAND;

    if (flags & EV_NOWAIT)
	{
	pSOSRunningTask->acceptEvents(eventsp,events);

	if (mode == pSOSFlagPendOR)
	    return *eventsp ? SUCCESS : ERR_NOEVS;

	return *eventsp == events ? SUCCESS : ERR_NOEVS;
	}
    
    if (pSOSRunningTask->pendEvents(events,mode,timeout) == CK_SYNCH_WAIT)
	{
	resched();
	    
	if (pSOSRunningTask->tstStatus(CKTIMEO))
	    {
	    *eventsp = pSOSRunningTask->flagValGet();
	    return ERR_TIMEOUT;
	    }
	}

    *eventsp = pSOSRunningTask->flagValGet();

    return SUCCESS;
}

int pSOSNode::kdoor(ascatch) (void (*asr)(cksigmask_t sigs),
			      ckflags_t mode)
{
    mustBeSynch();

    int il = (int)((mode >> 8) & 0x7);
    ckflags_t rtmode = psos2vrtosMode(mode);

    pSOSRunningTask->catchAsr(asr,rtmode,il);

    // CkNode::resched() will also check for pending signals.
    resched();

    return SUCCESS;
}

int pSOSNode::kdoor(assend) (CkId tid,
			     cksigmask_t sigs)
{
    if (tid.getType() != pSOS_TASK_INODE)
	return ERR_OBJTYPE;

    pSOSTask *task = (pSOSTask *)findActiveObject(tid);

    if (!task)
	return isDeletedObject(tid) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(tid);

    if (task->getAsrHandler() == CK_INVALID_ASR)
	return ERR_NOASR;

    if (!sigs)
	return SUCCESS;

    task->setSignals(sigs);

    CkNode *targetNode = CkNodeTable[tid.getNode()];
    targetNode->resched();

    return SUCCESS;
}

int pSOSNode::kdoor(asreturn) ()

{
    if (pSOSRunningTask->getAsrNestings() == 0)
	return ERR_NOTINASR;

    pSOSRunningTask->exitAsr();

    return SUCCESS;
}

int pSOSNode::kdoor(smcreate) (CkExtId uname,
			       ckulw_t count,
			       ckflags_t flags,
			       CkId *sema4idp)
{
    mustBeSynch();

    usema4++; // Track KC_NSEMA4 value.

    pSOSSem *sema4 = new pSOSSem(CkNameOf(uname),
				 uname,
				 count,
				 flags,
				 this);

    *sema4idp = sema4->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(smident) (CkExtId uname,
			      u_long nid,
			      CkId *sema4idp)
{
    mustBeSynch();

    if (checkpSOSNodeID(nid) < 0)
	return ERR_NODENO;

    CkObject *object = findActiveObject(uname,nid,pSOS_SEMA4_INODE);
	
    if (!object)
	return ERR_OBJNF;

    *sema4idp = object->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(smdelete) (CkId sema4id)

{
    mustBeSynch();

    if (sema4id.getType() != pSOS_SEMA4_INODE)
	return ERR_OBJTYPE;

    pSOSSem *sema4 = (pSOSSem *)findActiveObject(sema4id);

    if (!sema4)
	return isDeletedObject(sema4id) ? ERR_OBJDEL : ERR_OBJID;

    if (sema4id.getNode() != getNid())
	return ERR_ILLRSC;

    usema4--;

    int rc;

    if (sema4->getPendCount() > 0)
	rc = ERR_TATSDEL;
    else
	rc = SUCCESS;

    delete sema4;

    if (rc == ERR_TATSDEL)
	// Since remote tasks may be pending on this semaphore, start
	// the multi-processor rescheduling procedure.
	multiProcSched(pSOS_MAGIC);

    return rc;
}

int pSOSNode::kdoor(smp) (CkId sema4id,
			  ckflags_t flags,
			  ckticks_t timeout)
{
    if (sema4id.getType() != pSOS_SEMA4_INODE)
	return ERR_OBJTYPE;

    pSOSSem *sema4 = (pSOSSem *)findActiveObject(sema4id);

    if (!sema4)
	return isDeletedObject(sema4id) ? ERR_OBJDEL : ERR_OBJID;

    if (flags & SM_NOWAIT)
	{
	mustBeLocalOrSynch(sema4id);
	
	if (sema4->test() == CK_SYNCH_NOWAIT)
	    return ERR_NOSEM;
	}
    else
	{
	mustBeSynch();

	if (sema4->acquire(pSOSRunningTask,timeout) == CK_SYNCH_WAIT)
	    {
	    resched();

	    if (pSOSRunningTask->tstStatus(CKRMID))
		return ERR_SKILLD; // Semaphore deleted while pending.

	    if (pSOSRunningTask->tstStatus(CKTIMEO))
		return ERR_TIMEOUT; // Timeout.
	    }
	}

    return SUCCESS;
}

int pSOSNode::kdoor(smv) (CkId sema4id)

{
    if (sema4id.getType() != pSOS_SEMA4_INODE)
	return ERR_OBJTYPE;

    pSOSSem *sema4 = (pSOSSem *)findActiveObject(sema4id);

    if (!sema4)
	return isDeletedObject(sema4id) ? ERR_OBJDEL : ERR_OBJID;

    mustBeLocalOrSynch(sema4id);
    
    if (sema4->release() == CK_SYNCH_RESCHED)
	multiProcSched(pSOS_MAGIC);

    return SUCCESS;
}

int pSOSNode::kdoor(smav) (CkId sema4id)

{
    mustBeSynch();
    
    int rc = kdoor(smv)(sema4id);

    if (rc != SUCCESS)
	dispatchAsyncErr(rc);

    return SUCCESS;
}

int pSOSNode::kdoor(tmset) (ckulw_t date,
			    ckulw_t time,
			    ckticks_t ticks)
{
    mustBeIsrOrSynch();
    
    CkDateVal dv(date,time,ticks,getTicks2Sec());

    if (dv.getMonth() < 1 ||
	dv.getMonth() > 12 ||
	dv.getDay() < 1 ||
	dv.getDay() > 31)
	return ERR_ILLDATE;

    if (dv.getHour() > 23 ||
	dv.getMin() > 59 ||
	dv.getSec() > 59)
        return ERR_ILLTIME;

    if (dv.getTicks() >= getTicks2Sec())
	return ERR_ILLTICKS;

    setNodeDate(dv);

    resched();

    return SUCCESS;
}

int pSOSNode::kdoor(tmget) (ckulw_t *datep,
			    ckulw_t *timep,
			    ckulw_t *ticksp)
{
    if (!tstStatus(CKTMSET))
	return ERR_NOTIME;	// Must call tm_set() first.

    CkDateVal dv = getNodeDate();

    *datep = dv.getDate();
    *timep = dv.getTime();
    *ticksp = dv.getTicks();

    return SUCCESS;
}

void pSOSNode::kdoor(tmwkafter) (ckticks_t ticks)

{
    mustBeSynch();

    if (ticks > 0)
	suspendThread(pSOSRunningTask,CKDELAY,ticks);
    else
	readyThread(pSOSRunningTask); // Perform manual round-robin.

    resched();
}

int pSOSNode::kdoor(tmwkwhen) (ckulw_t date,
			       ckulw_t time,
			       ckticks_t ticks)
{
    mustBeSynch();

    if (!tstStatus(CKTMSET))
	return ERR_NOTIME;	// Must call tm_set() first.

    CkDateVal dv(date,time,ticks,getTicks2Sec());

    if (dv.getMonth() < 1 ||
	dv.getMonth() > 12 ||
	dv.getDay() < 1 ||
	dv.getDay() > 31)
	return ERR_ILLDATE;

    if (dv.getHour() > 23 ||
	dv.getMin() > 59 ||
	dv.getSec() > 59)
        return ERR_ILLTIME;

    if (dv.getTicks() >= getTicks2Sec())
	return ERR_ILLTICKS;

    if (dv <= getNodeDate())
	return ERR_TOOLATE;

    pSOSAlarm *alarm = new pSOSAlarm(pSOSRunningTask,dv);

    scheduleAlarm(alarm);

    suspendThread(pSOSRunningTask,CKDELAY);

    resched();

    return SUCCESS;
}

int pSOSNode::kdoor(tmevafter) (ckticks_t ticks,
				ckflags_t events,
				CkId *timeridp)
{
    mustBeSynch();

    utimer++;	// Track KC_NTIMER value.

    pSOSTimer *evtm = new pSOSTimer(pSOSRunningTask,
				    ticks,
				    0,
				    events);
    scheduleAlarm(evtm);

    *timeridp = evtm->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(tmevwhen) (ckulw_t date,
			       ckulw_t time,
			       ckticks_t ticks,
			       ckflags_t events,
			       CkId *timeridp)
{
    mustBeSynch();

    if (!tstStatus(CKTMSET))
	return ERR_NOTIME;	// Must call tm_set() first.

    CkDateVal dv(date,time,ticks,getTicks2Sec());

    if (dv.getMonth() < 1 ||
	dv.getMonth() > 12 ||
	dv.getDay() < 1 ||
	dv.getDay() > 31)
	return ERR_ILLDATE;

    if (dv.getHour() > 23 ||
	dv.getMin() > 59 ||
	dv.getSec() > 59)
        return ERR_ILLTIME;

    if (dv.getTicks() >= getTicks2Sec())
	return ERR_ILLTICKS;

    if (dv <= getNodeDate())
	return ERR_TOOLATE;

    utimer++;	// Track KC_NTIMER value.

    pSOSTimer *evtm = new pSOSTimer(pSOSRunningTask,
				    dv,
				    events);
    scheduleAlarm(evtm);

    *timeridp = evtm->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(tmevevery) (ckticks_t ticks,
				ckflags_t events,
				CkId *timeridp)
{
    mustBeSynch();

    utimer++;	// Track KC_NTIMER value.

    pSOSTimer *evtm = new pSOSTimer(pSOSRunningTask,
				    ticks,
				    ticks,
				    events);
    scheduleAlarm(evtm);

    *timeridp = evtm->getCkId();

    return SUCCESS;
}

int pSOSNode::kdoor(tmcancel) (CkId timerid)

{
    mustBeSynch();
		 
    if (timerid.getType() != pSOS_TIMER_INODE)
	return ERR_BADTMID;

    CkAlarm *timer = (CkAlarm *)findActiveObject(timerid);

    if (!timer)
	return isDeletedObject(timerid) ? ERR_TMNOTSET : ERR_BADTMID;

    cancelAlarm(timer);	// This deletes the alarm object.

    return SUCCESS;
}

void pSOSNode::kdoor(kfatal) (ckulw_t ecode,
			      ckulw_t flags)
{
    mustBeInitOrSynch();

    if ((flags & K_GLOBAL) || CkNodeList.getCount() == 1)
	{
	if (fatalHandler)
	    {
	    fatalHandler(ecode);
	    _exit(ecode);
	    }

	CkManager::This->fatal("fatal error: node=%d, status=%x\n",
			       getNid(),
			       ecode);
	}

    if (fatalHandler)
	fatalHandler(ecode);
    else
	CkManager::This->warning("fatal error: node=%d, status=%x\n",
				 getNid(),
				 ecode);

    hold(CKHALT); // Never returns.
}

int pSOSNode::kdoor(kterminate) (u_long nid,
				 ckulw_t fcode,
				 ckulw_t flags)
{
    if (nid <= 1)
	return ERR_MASTER;

    if (nid > CkNodeList.getCount() ||
	CkNodeTable[nid - 1]->getNodeType() != pSOS_MAGIC)
	return ERR_NODENO;

    pSOSNode *targetNode = (pSOSNode *)CkNodeTable[nid - 1];
    targetNode->kdoor(kfatal)(fcode,K_LOCAL);
    
    return SUCCESS;
}

// Export Protocol object

pSOSNodeExportMessage::pSOSNodeExportMessage (pSOSNode *node) :
    CkNodeExportMessage(node->getName())
{
    nid = node->getNid();
    scopy(osName,node->getOS(),sizeof(osName) - 1);
    strcpy(threadTypeName,"Task");
    asciiXID = True;
    nThreadStates = 5;
    strcpy(threadStates[0].stateLabel,"Halted");
    threadStates[0].stateHandle = pSOSTaskHALTED;
    strcpy(threadStates[1].stateLabel,"Dormant");
    threadStates[1].stateHandle = pSOSTaskDORMANT;
    strcpy(threadStates[2].stateLabel,"Blocked");
    threadStates[2].stateHandle = pSOSTaskBLOCKED;
    strcpy(threadStates[3].stateLabel,"Ready");
    threadStates[3].stateHandle = pSOSTaskREADY;
    strcpy(threadStates[4].stateLabel,"Executing");
    threadStates[4].stateHandle = pSOSTaskEXECUTING;
}

// pSOSInterrupt -- A pSOS extension of the VRTOS interrupt object
// used by the trampoline code to map the pSOS semantics to
// Ck's.

extern "C" {

static int psos_isr_trampoline (ckvector_t vector,
				void *cookie)
{
    pSOSInterrupt *intr = (pSOSInterrupt *)cookie;
    intr->routine();
    intr->setStatus(CK_INTR_ACKED);

    return CK_ISR_HANDLED;
}

}

pSOSInterrupt::pSOSInterrupt (ckvector_t _vector,
			      cklevel_t _level,
			      void (*_routine)(void),
			      const char *_name,
			      pSOSNode *_node) :
    CkInterrupt(_vector,
		_level,
		psos_isr_trampoline,
		NULL,
		this,
		_name,
		_node),
    routine(_routine) {
}
