/*
 * This file is part of the pSOS simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the pSOS+m kernel synchronization objects
 * such as semaphores, event flag groups and message queues.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "CkSynch.h"
#include "pSOSNode.h"
#include "pSOSTask.h"
#include "pSOSSynch.h"

pSOSSynch::pSOSSynch(const char *_name,
		     InsertMode _mode,
		     int _options,
		     CkStateExportObject *_synchState,
		     CkNode *_node) :
    CkSynchro(_name,
	      _mode,
	      _synchState,
	      _node) {
    options = _options;
}

void pSOSSynch::blockTask (pSOSTask *task, ckticks_t timeout)

{
    // If target task's node is remote, get an RPC agent.

    if (task->getNode() != getNode())
	task->getNode()->getRPCAgent(getCkId(),
					   task->getCkId());

    insert(task,task->getPri());
    task->clrStatus(CKTIMEO|CKRMID);
    task->getNode()->suspendThread(task,CKPEND,timeout,this);
}

void pSOSSynch::unblockTask (pSOSTask *task)

{
    if (task->getNode() != getNode())
	task->getNode()->freeRPCAgent();

    task->getNode()->readyThread(task,CKPEND);
}

void pSOSSynch::forgetTask (pSOSTask *task)

{
    CkSynchro::forget(task);

    if (task->getNode() != getNode())
	task->getNode()->freeRPCAgent();
}

const char *pSOSSem::stateArray[3] = {
    "IDLE",
    "PENDED",
    "POSTED"
};

pSOSSem::pSOSSem (const char *_name,
		  CkExtId _uname,
		  ckulw_t _icount,
		  int _options,
		  pSOSNode *_node) :
    pSOSSynch(_name,
	      (_options & SM_PRIOR) ? PRDN : FIFO,
	      _options,
	      new CkStateExportObject(_node,
				      _name,
				      "Semaphores",
				      stateArray,3),
	      _node)
{
    count = _icount;

    synchState->setState(count == 0 ? pSOSSemIDLE : pSOSSemPOSTED);

    _node->nameObject(this,
		      pSOS_SEMA4_INODE,
		      testOption(SM_GLOBAL) ? CkGlobalDomain : CkLocalDomain,
		      _uname);

    protoInit();
}

pSOSSem::~pSOSSem ()

{
    pSOSTask *task;

    while ((task = (pSOSTask *)get()) != NULL)
	{
	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void pSOSSem::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("psos","pSOSSem"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

const char *pSOSSem::getTypedName ()

{
    CkSynchro::typedName.set("sema4");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void pSOSSem::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("value");
    _tclist.append(count);
    
    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void pSOSSem::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"value") == 0)
	    {
	    ckusw_t newCount = (ckusw_t)atoi(parser.next());

	    if (newCount > count)
		{
		for (ckusw_t relCount = newCount - count;
		     relCount > 0; relCount--)
		    release();
		}
	    else
		count = newCount;
	    }
	}
}

int pSOSSem::acquire (pSOSTask *task, ckticks_t timeout)

{
    if (count > 0)
	{
	if (--count == 0)
	    synchState->setState(pSOSSemIDLE);

	return CK_SYNCH_DONE;
	}

    blockTask(task,timeout);
    synchState->setState(pSOSSemPENDED);

    return CK_SYNCH_WAIT;
}

int pSOSSem::release ()

{
    pSOSTask *task = (pSOSTask *)get();

    if (task)
	{
	unblockTask(task);

	if (getCount() == 0)
	    synchState->setState(pSOSSemIDLE);

	return CK_SYNCH_RESCHED;
	}

    if (count + 1 != 0)		// Check for roll-over.
	{
	count++;
	synchState->setState(pSOSSemPOSTED);
	}

    return CK_SYNCH_DONE;
}

int pSOSSem::test ()

{
    if (count > 0)
	{
	if (--count == 0)
	    synchState->setState(pSOSSemIDLE);

	return CK_SYNCH_DONE;
	}

    return CK_SYNCH_NOWAIT;
}

void pSOSSem::forget (CkThread *thread)

{
    forgetTask((pSOSTask *)thread);

    if (getCount() == 0)
	synchState->setState(pSOSSemIDLE);
}

const char *pSOSQueue::stateArray[5] = {
    "EMPTY",
    "PENDED",
    "POSTED",
    "FULL",
    "JAMMED"
};

pSOSQueue::pSOSQueue (const char *_name,
		      CkExtId _uname,
		      ckulw_t _qsize,
		      ckulw_t _msize,
		      int _options,
		      pSOSNode *_node) :
    pSOSSynch(_name,
	      (_options & Q_PRIOR) ? PRDN : FIFO,
	      _options,
	      new CkStateExportObject(_node,
				      _name,
				      "Queues",
				      stateArray,5),
	      _node)
{
    // qize == -1 means `no limit', qsize == 0 means that it may
    // be posted only if already pended by another task.

    qsize = _qsize;
    msize = _msize;

    synchState->setState(pSOSQueueEMPTY);

    _node->nameObject(this,
		      pSOS_QUEUE_INODE,
		      testOption(Q_GLOBAL) ? CkGlobalDomain : CkLocalDomain,
		      _uname);

    protoInit();
}

pSOSQueue::~pSOSQueue ()

{
    if (!testOption(Q_PRIBUF))
	{
	if (qsize == (ckulw_t)~0)
	    getNode()->releaseMsgBufs(inq.getCount());
	else
	    getNode()->releaseMsgBufs(qsize);
	}

    pSOSTask *task;

    while ((task = (pSOSTask *)get()) != NULL)
	{
	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void pSOSQueue::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("psos","pSOSQueue"));
    protoExport(&em,sizeof(em));
    synchState->protoInit();
}

void pSOSQueue::dynamicExpose ()

{
    TclList ___tclist, __tclist, _tclist, tclist;

    _tclist.set("type");
    _tclist.append(testOption(Q_VARIABLE) ? "variable" : "fixed");
    tclist.append(_tclist);
	     
    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);
    __tclist.clear();

    _tclist.set("messages");

    for (pSOSMessage *msg = inq.first(); msg; msg = msg->link)
	{
	___tclist.clear();

	if (testOption(Q_VARIABLE))
	    {
	    int nbytes = Min(16,msg->length);
	    TclList ___tclist;

	    for (char *dstart = msg->u.data,
		     *dend = msg->u.data + nbytes;
		 dstart < dend; dstart++)
		___tclist.appendx((u_long)*dstart);

	    ___tclist.append(CString().format("(sz=%u)",msg->length));
	    }
	else
	    {
	    for (int n = 0; n < 4; n++)
		___tclist.appendx(msg->u.fixbuf[n]);
	    }

	___tclist.append(msg->time.format(NULL));
	__tclist.append(___tclist);
	}
    
    _tclist.append(__tclist);
    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void pSOSQueue::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	// Only fixed-size queues can be altered from
	// the ISE by adding messages.

	if (!testOption(Q_VARIABLE) && strcmp(setting,"messages") == 0)
	    {
	    TclListParser _parser(parser.next());
	    const char *values;
	    int posted = 0;

	    while ((values = _parser.next()) != NULL)
		{
		TclListParser __parser(values);
		u_long m[4];
		CString cs;

		for (int n = 0; n < 4; n++)
		    m[n] = cs.overwrite(__parser.next()).getHex();
		
		if (post(m,sizeof(m)) <= 0)
		    break;

		posted++;
		}
	    }
	}
}

const char *pSOSQueue::getTypedName ()

{
    CkSynchro::typedName.set("queue");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

pSOSMessage *pSOSQueue::buildMessage (void *buf, ckulw_t length)

{
    pSOSMessage *msg;
    ckulw_t nbytes = sizeof(*msg);

    if (length > sizeof(msg->u))
	nbytes += (length - sizeof(msg->u));

    msg = (pSOSMessage *)new char[nbytes];
    msg->time = SxClock;
    msg->length = length;
    memcpy(msg->u.data,buf,length);

    return msg;
}

int pSOSQueue::pend (pSOSTask *task, ckticks_t timeout)

{
    pSOSMessage *msg = inq.pop();

    if (msg)
	{
	task->msgSet(msg);

	if (!testOption(Q_PRIBUF))
	    getNode()->releaseMsgBufs(1);

	synchState->setState(inq.getCount() == 0 ?
			     pSOSQueueEMPTY : pSOSQueuePOSTED);

	return CK_SYNCH_DONE;
	}
    
    blockTask(task,timeout);

    if (getCount() == 0)
	synchState->setState(pSOSQueuePENDED);

    return CK_SYNCH_WAIT;
}

int pSOSQueue::post (void *buf, ckulw_t length)

{
    pSOSTask *task = (pSOSTask *)get();
    
    if (task)
	{
	task->msgSet(buildMessage(buf,length));

	unblockTask(task);

	if (getCount() == 0)
	    synchState->setState(pSOSQueueEMPTY);

	return CK_SYNCH_RESCHED;
	}

    if (inq.getCount() >= qsize)
	return CK_SYNCH_ERROR;	// Queue is full.

    if (!testOption(Q_PRIBUF) && !getNode()->holdMsgBufs(1))
	return CK_SYNCH_NOWAIT; // No more buffers in the shared pool.

    inq.append(buildMessage(buf,length));

    synchState->setState(inq.getCount() >= qsize ?
			 pSOSQueueFULL : pSOSQueuePOSTED);

    return CK_SYNCH_DONE;
}

int pSOSQueue::jam (void *buf, ckulw_t length)

{
    pSOSTask *task = (pSOSTask *)get();
    
    if (task)
	{
	task->msgSet(buildMessage(buf,length));

	unblockTask(task);

	if (getCount() == 0)
	    synchState->setState(pSOSQueueEMPTY);

	return CK_SYNCH_RESCHED;
	}

    if (inq.getCount() >= qsize)
	return CK_SYNCH_ERROR;	// Queue is full.

    if (!testOption(Q_PRIBUF) && !getNode()->holdMsgBufs(1))
	return CK_SYNCH_NOWAIT; // No more buffers in the shared pool.

    inq.prepend(buildMessage(buf,length));

    synchState->setState(inq.getCount() >= qsize ?
			 pSOSQueueFULL : pSOSQueuePOSTED);

    return CK_SYNCH_DONE;
}

pSOSMessage *pSOSQueue::accept ()

{
    pSOSMessage *msg = inq.pop();

    if (msg)
	{
	if (!testOption(Q_PRIBUF))
	    getNode()->releaseMsgBufs(1);

	synchState->setState(inq.getCount() == 0 ?
			     pSOSQueueEMPTY : pSOSQueuePOSTED);

	return msg;
	}
    
    return NULL;
}

ckulw_t pSOSQueue::broadcast (void *buf, ckulw_t length)

{
    ckulw_t qpostable = Min(getPendCount(),qsize - getMsgCount());
    ckulw_t qposted = 0;

    while (qpostable-- > 0)
	{
	post(buf,length); // Cannot fail.
	qposted++;
	}
	    
    return qposted;
}

void pSOSQueue::forget (CkThread *thread)

{
    forgetTask((pSOSTask *)thread);

    if (getCount() == 0)
	synchState->setState(pSOSQueueEMPTY);
}

pSOSFlagGroup::pSOSFlagGroup (const char *_name,
			      CkExtId _uname,
			      InsertMode _mode,
			      CkDomain _domain,
			      pSOSNode *_node) :
    pSOSSynch(_name,
	      _mode,
	      0,
	      NULL,
	      _node)
{
    fgroup = 0;

    _node->nameObject(this,pSOS_EVTGRP_INODE,_domain,_uname);

    protoInit();
}

pSOSFlagGroup::~pSOSFlagGroup ()

{
    pSOSTask *task;

    while ((task = (pSOSTask *)get()) != NULL)
	{
	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void pSOSFlagGroup::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("psos","pSOSFlagGroup"));

    protoExport(&em,sizeof(em));
}

void pSOSFlagGroup::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("flags");
    __tclist.clear();
    ckflags_t flags = fgroup;

    for (int n = 0; n < 32; n++)
	{
	__tclist.append(flags & 1);
	flags >>= 1;
	}

    _tclist.append(__tclist);
    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void pSOSFlagGroup::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"flags") == 0)
	    {
	    TclListParser _parser(parser.next());
	    const char *flag;
	    ckflags_t flags = 0;

	    // Do reverse scanning (bit 31->0).
	    _parser.reset(True);

	    while ((flag = _parser.prev()) != NULL)
		{
		flags |= atoi(flag);
		flags <<= 1;
		}

	    fgroup = 0;
	    post(flags);
	    }
	}
}

const char *pSOSFlagGroup::getTypedName ()

{
    CkSynchro::typedName.set("event flags");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

int pSOSFlagGroup::pend (pSOSTask *task,
			 ckflags_t mask,
			 pSOSFlagPendModes mode,
			 ckticks_t timeout)
{
    if (!mask)
	{
	// As pSOS documentation states it, a null mask
	// causes the pending events to be returned but
	// not cleared.
	task->flagValSet(fgroup);
	return CK_SYNCH_DONE;
	}

    if (mode == pSOSFlagPendOR)
	{
	if (mask & fgroup)
	    {
	    task->flagValSet(fgroup & mask);
	    fgroup &= ~mask;
	    return CK_SYNCH_DONE;
	    }
	}
    else if (mode == pSOSFlagPendAND)
	{
	if ((fgroup & mask) == mask)
	    {
	    task->flagValSet(mask);
	    fgroup &= ~mask;
	    return CK_SYNCH_DONE;
	    }
	}

    task->flagModeSet(mode);
    task->flagValSet(mask);
    blockTask(task,timeout);

    return CK_SYNCH_WAIT;
}

int pSOSFlagGroup::post (ckflags_t mask)

{
    int readied = 0;

    fgroup |= mask;

    pSOSTaskIterator it(this);
    pSOSTask *task, *ntask = it.next();

    while ((task = ntask) != NULL)
	{
	pSOSFlagPendModes mode = task->flagModeGet();
	ckflags_t wmask = task->flagValGet();

	if (mode == pSOSFlagPendOR)
	    {
	    if (wmask & fgroup)
		{
		ntask = it.pop();
		task->flagValSet(fgroup & wmask);
		fgroup &= ~wmask;
		unblockTask(task);
		readied++;
		continue;
		}
	    }
	else if (mode == pSOSFlagPendAND)
	    {
	    if ((fgroup & wmask) == wmask)
		{
		ntask = it.pop();
		task->flagValSet(wmask);
		fgroup &= ~wmask;
		unblockTask(task);
		readied++;
		continue;
		}
	    }

	ntask = it.next();
	}

    return readied > 0 ? CK_SYNCH_RESCHED : CK_SYNCH_DONE;
}

void pSOSFlagGroup::accept (ckflags_t *rflags, ckflags_t mask)

{
    ckflags_t oflags = (fgroup & mask);
    fgroup &= ~mask;
    *rflags = oflags;
}

void pSOSFlagGroup::forget (CkThread *thread)

{
    pSOSTask *task = (pSOSTask *)thread;

    forgetTask(task);

    // Save the currently pending events. They will be
    // returned to the task along with the timeout error.
    
    task->flagValSet(fgroup);
}
