/*
 * This file is part of the pSOS simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the pSOS task.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include "CString.h"
#include "CkManager.h"
#include "pSOSProto.h"
#include "pSOSNode.h"
#include "pSOSTask.h"

const char *pSOSTask::stateArray[5] = {

    "HALTED",
    "DORMANT",
    "BLOCKED",
    "READY",
    "RUNNING"
};

pSOSTask::pSOSTask (const char *_name,
		    CkExtId _uname,
		    int _priority,
		    CkDomain _domain,
		    pSOSNode *_node,
		    unsigned _stackSize) :
    CkThread(_node,
	     _name,
	     _priority,
	     new CkThreadState(_name,
			       "Tasks",
			       pSOSTask::stateArray,5,
			       _node,
			       this),
	     Max(_stackSize,16384)) // Enforce a minimum 16Kb user stack.
{
    lastStateChangeTime = 0;

    stateDistribs[0] = new CkHistogram(_name,"Blocked state",_node);
    stateDistribs[0]->protoInit();
    stateDistribs[1] = new CkHistogram(_name,"Ready state",_node);
    stateDistribs[1]->protoInit();
    stateDistribs[2] = new CkHistogram(_name,"Executing state",_node);
    stateDistribs[2]->protoInit();

    eventGroup = new pSOSFlagGroup(CkNameOf(_uname),
				   "EVNT",
				   PRDN,
				   CkLocalDomain,
				   _node);

    memset(notepad,0,sizeof(notepad));

    _node->nameObject(this,pSOS_TASK_INODE,_domain,_uname);

    _node->declareThread(this);

    tstate->protoInit();
}

pSOSTask::~pSOSTask ()

{
    delete eventGroup;
    getNode()->unnameObject(this);
}

void pSOSTask::start (void (*_startAddr)(u_long,u_long,u_long,u_long),
		      ckflags_t _mode,
		      int _il,
		      pSOSArgs *_targs)
{
    startAddr = _startAddr;
    startPri = getPri();
    startMode = _mode;
    startTaskIL = _il;

    if (_targs)
	startArgs = *_targs;
    else
	memset(startArgs,0,sizeof(startArgs));

    protoInit();

    getNode()->startThread(this,_mode,_il);
}

void pSOSTask::restart (pSOSArgs *_targs)

{
    if (_targs)
	startArgs = *_targs;
    else
	memset(startArgs,0,sizeof(startArgs));

    // Break the thread out of any wait it is currently in.
    getNode()->unblockThread(this);

    // If the task has been explicitely suspended, resume it.
    if (tstStatus(CKSUSP))
	getNode()->readyThread(this,CKSUSP);

    // Reset modebits.
    clrStatus(CK_MODE_BITS);
    setStatus(startMode);

    // Reset task priority to the initial one.
    setPri(startPri);

    // Clear pending signals.
    clrSignals();

    // Reset task's interrupt mask.
    setThreadIL(startTaskIL);

    // Cancel pending alarms for this task.
    getNode()->cancelClientAlarms(this);

    if (tstStatus(pSOS_TASK_STARTED))
	{
	resetTracking();
	xraise(restartContext);
	}
}

void pSOSTask::body () {

    kroot(run)();
}

void pSOSTask::kroot(run) ()

{
    xtry(restartContext);

    getNode()->reinstateThread(this);

    setStatus(pSOS_TASK_STARTED);

    startAddr(startArgs[0],
	      startArgs[1],
	      startArgs[2],
	      startArgs[3]);

    CkManager::This->warning("task `%s' should call t_delete(0) to terminate",
			     (const char *)CkNameOf(this));

    getNode()->kdoor(tdelete)(0);
}

void pSOSTask::startAsr (cksigmask_t sigs)

{
    if (raiseAsr(sigs) == 0)
	CkManager::This->warning("exiting ASR didn't call as_return()");
}

void pSOSTask::logTaskState (pSOSTaskStates s)

{
    int olds = getThreadState();

    ITime dtLap = SxClock - lastStateChangeTime;

    if (olds >= 2 && olds <= 4)
	// <!> Depends on pSOSTaskStates enum order.
	stateDistribs[olds - 2]->add(dtLap);

    lastStateChangeTime = SxClock;

    setThreadState(s);
}

void pSOSTask::setReady () {

    logTaskState(pSOSTaskREADY);
}

void pSOSTask::setSuspended () {

    logTaskState(pSOSTaskBLOCKED);
}

void pSOSTask::setPending () {

    setSuspended();
}

void pSOSTask::setRunning () {

    logTaskState(pSOSTaskEXECUTING);
}

void pSOSTask::setDelayed () {

    setSuspended();
}

void pSOSTask::setHalted () {
    
    logTaskState(pSOSTaskHALTED);
}

void pSOSTask::setDormant () {

    logTaskState(pSOSTaskDORMANT);
}

void (*pSOSTask::getStartAddr ())(...) {

    return (void (*)(...))startAddr;
}

void pSOSTask::protoInit ()

{
    CkThreadExportMessage tem(this,"psos","pSOSTask");
    protoExport(&tem,sizeof(tem));
}

void pSOSTask::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("prio");
    _tclist.append(getPri());
    tclist.append(_tclist);

    _tclist.set("ilevel");
    _tclist.append(getThreadIL());
    tclist.append(_tclist);

    _tclist.set("signals");
    __tclist.clear();

    cksigmask_t sigmask = getSignals();

    for (int n = 0; n < 32; n++)
	{
	__tclist.append(sigmask & 1);
	sigmask >>= 1;
	}
    
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("notepad");
    __tclist.clear();

    for (int n = 0; n < pSOS_NOTEPAD_SIZE; n++)
	__tclist.appendx(getRegister(n));
    
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("state");
    _tclist.append(stateArray[getThreadState()]);
    tclist.append(_tclist);

    if (tstStatus(CKPEND))
	{
	// The wait channel is the name of the resource on which
	// the task is currently pending.
	_tclist.set("wchan");
	_tclist.append(getPendedSynchro()->first()->getTypedName());
	tclist.append(_tclist);
	}
    
    if (tstStatus(CKDELAY))
	{
	_tclist.set("timeout");
	_tclist.append(getTimerCount());
	tclist.append(_tclist);
	}

    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void pSOSTask::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"prio") == 0)
	    {
	    int prio = atoi(parser.next());
	    getNode()->reniceThread(this,prio);
	    }
	else if (strcmp(setting,"ilevel") == 0)
	    {
	    int ilevel = atoi(parser.next());
	    setThreadIL(ilevel);
	    }
	else if (strcmp(setting,"signals") == 0 &&
		 getAsrHandler() != CK_INVALID_ASR)
	    {
	    TclListParser _parser(parser.next());
	    const char *sigstate;
	    cksigmask_t sigmask = 0;

	    // Do reverse scanning (bit 31->0).
	    _parser.reset(True);

	    while ((sigstate = _parser.prev()) != NULL)
		{
		sigmask |= atoi(sigstate);
		sigmask <<= 1;
		}

	    setSignals(sigmask);
	    }
	}
}

pSOSAlarm::pSOSAlarm (pSOSTask *_task, CkDateVal _dv) :
    CkAlarm(_task,_dv) {

    clientTask = _task;
}

void pSOSAlarm::timeout (CkObject *client) {

    // Attempt to wake the task up removing the delay condition.
    clientTask->getNode()->readyThread(clientTask,CKDELAY);
}

pSOSTimer::pSOSTimer (pSOSTask *_task,
		      CkDateVal _dv,
		      ckflags_t _events) :
    CkAlarm(_task,_dv)
{
    clientTask = _task;
    events = _events;

    _task->getNode()->nameObject(this,
				 pSOS_TIMER_INODE,
				 CkLocalDomain,
				 "EVTM");
}

pSOSTimer::pSOSTimer (pSOSTask *_task,
		      ckticks_t _initialTicks,
		      ckticks_t _reloadTicks,
		      ckflags_t _events) :
    CkAlarm(_task,
	    _initialTicks,
	    _reloadTicks)
{
    clientTask = _task;
    events = _events;

    _task->getNode()->nameObject(this,
				 pSOS_TIMER_INODE,
				 CkLocalDomain,
				 "EVTM");
}

pSOSTimer::~pSOSTimer () {

    ((pSOSNode *)getNode())->unnameObject(this);
}

void pSOSTimer::timeout (CkObject *client) {

    // Send programmed events to the task.
    clientTask->postEvents(events);
}
